"""
Book tests.
"""

import textwrap
from lute.models.book import Book


def test_create_book(english):
    """
    When a book is created with given content, the content
    is split into separate Text objects.
    """
    fulltext = "Here is a dog. And a cat."
    scenario(english, fulltext, 5, ["Here is a dog.[4]", "And a cat.[3]"])
    scenario(english, fulltext, 500, ["Here is a dog. And a cat.[7]"])
    scenario(
        english,
        fulltext + " And a thing.",
        8,
        ["Here is a dog. And a cat.[7]", "And a thing.[3]"],
    )
    scenario(
        english,
        "Here is a dog.\nAnd a cat.",
        500,
        ["Here is a dog.\nAnd a cat.[7]"],
    )
    scenario(
        english,
        "Here is a dog.\n\n\nAnd a cat.",
        500,
        ["Here is a dog.\n\n\nAnd a cat.[7]"],
    )


def test_create_book_page_breaks(english):
    "--- on its own line means a page break."
    fulltext = textwrap.dedent(
        """\
      Here is a dog.
      And a cat.
    """
    )
    scenario(english, fulltext, 500, ["Here is a dog.\nAnd a cat.[7]"])

    fulltext = textwrap.dedent(
        """\
      Here is a dog.
      ---
      And a cat.
    """
    )
    scenario(english, fulltext, 500, ["Here is a dog.[4]", "And a cat.[3]"])

    fulltext = textwrap.dedent(
        """\
      Here is a dog.
      ---
      ---
      ---
      ---
      And a cat.
      ---
      ---
    """
    )
    scenario(english, fulltext, 500, ["Here is a dog.[4]", "And a cat.[3]"])


def scenario(english, fulltext, maxwords, expected):
    """
    Check scenarios.
    """
    b = Book.create_book("hi", english, fulltext, maxwords)

    actuals = [f"{t.text}[{t.word_count}]" for t in b.texts]
    print(actuals)
    assert "/".join(actuals) == "/".join(expected), f"scen {maxwords}, {fulltext}"



"Smoke test only."

from lute.cli.language_term_export import generate_file


def test_smoke_test(app_context, tmp_path):
    "dump data."
    outfile = tmp_path / "outfile.csv"
    generate_file("English", outfile)
    with open(outfile, "r", encoding="utf-8") as ofhandle:
        text = ofhandle.read()
    print(text)
    lines = text.split("\n")
    assert (
        lines[0] == "term,count,familycount,books,definition,status,children"
    ), "headings"
    assert lines[1].startswith("the,"), "the is most common"



"""
Language model tests - getting, saving, etc.

Low value but ensure that the db mapping is correct.
"""

from lute.models.language import Language
from tests.dbasserts import assert_sql_result


def test_demo_has_preloaded_languages(app_context):
    """
    When users get the initial demo, it has English, French, etc,
    pre-defined.
    """
    sql = """
    select LgName
    from languages
    where LgName in ('English', 'French')
    order by LgName
    """
    assert_sql_result(sql, ["English", "French"], "sanity check loaded")


def test_new_language_has_sane_defaults():
    """
    Only validates the call to __init__.  Sqlalchemy mappings aren't used during the constuctor.
    """
    lang = Language()
    assert lang.character_substitutions == "´='|`='|’='|‘='|...=…|..=‥"
    assert lang.regexp_split_sentences == ".!?"
    assert lang.exceptions_split_sentences == "Mr.|Mrs.|Dr.|[A-Z].|Vd.|Vds."
    assert lang.word_characters == "a-zA-ZÀ-ÖØ-öø-ȳáéíóúÁÉÍÓÚñÑ"
    assert lang.right_to_left is False
    assert lang.show_romanization is False
    assert lang.parser_type == "spacedel"


def test_can_find_lang_by_name(app_context):
    """
    Returns lang if found, or None
    """
    e = Language.find_by_name("English")
    assert e.name == "English", "case match"

    e_lc = Language.find_by_name("english")
    assert e_lc.name == "English", "case-insensitive"

    nf = Language.find_by_name("notfound")
    assert nf is None, "not found"


def test_language_word_char_regex_returns_python_compatible_regex(app_context):
    """
    Old Lute v2 ran in php, so the language word chars regex
    could look like this:

    x{0600}-x{06FF}x{FE70}-x{FEFC}  (where x = backslash-x)

    This needs to be converted to the python equivalent, e.g.

    u0600-u06FFuFE70-uFEFC  (where u = backslash-u)
    """
    a = Language.find_by_name("Arabic")
    assert a.word_characters == r"\u0600-\u06FF\uFE70-\uFEFC"


def test_lang_to_dict_from_dict_returns_same_thing(app_context):
    """
    Lang can be exported to yaml and imported from yaml.
    A dictionary is used as the intermediary form, so the
    same language should return the same data.
    """
    e = Language.find_by_name("English")
    e_dict = e.to_dict()
    e_from_dict = Language.from_dict(e_dict)
    e_back_to_dict = e_from_dict.to_dict()
    assert e_dict == e_back_to_dict, "Same thing returned"



"""
Common fixtures used by many tests.
"""

import os
import yaml
import pytest

from lute.config.app_config import AppConfig
from lute.db import db
from lute.language.service import get_language_def
import lute.db.demo
from lute.app_factory import create_app

from lute.models.language import Language


def pytest_sessionstart(session):  # pylint: disable=unused-argument
    """
    Ensure test config defines a test environment.

    Special configuration for test runs required:

    DBNAME must start with test_

    DATAPATH must be specified: this ensures that the tests don't
    accidentally write into the user_data (which could mess with prod
    data/media etc)
    """
    thisdir = os.path.dirname(os.path.realpath(__file__))
    configfile = os.path.join(thisdir, "..", "lute", "config", "config.yml")

    config = None
    with open(configfile, "r", encoding="utf-8") as f:
        config = yaml.safe_load(f)

    failures = []
    if "DATAPATH" not in config:
        failures.append("DATAPATH not in config file")

    ac = AppConfig(configfile)

    if not ac.is_test_db:
        failures.append("DBNAME in config.yml must start with test_")
    if len(failures) > 0:
        msg = f"Bad config.yml: {', '.join(failures)}"
        pytest.exit(msg)


@pytest.fixture(name="testconfig")
def fixture_config():
    "Config using the app config."
    ac = AppConfig(AppConfig.default_config_filename())
    yield ac


@pytest.fixture(name="app")
def fixture_app():
    """
    A clean instance of the demo database.

    I'm not a _huge_ fan of this because if the app
    is open while tests are running, the app seems to hold
    on to references to the old deleted db ...
    that said, it's much faster to do this than to do a
    "wipe and reload database" on every test run.
    """
    config_file = AppConfig.default_config_filename()
    c = AppConfig(config_file)
    if os.path.exists(c.dbfilename):
        os.unlink(c.dbfilename)
    extra_config = {"WTF_CSRF_ENABLED": False, "TESTING": True}
    app = create_app(config_file, extra_config=extra_config)
    yield app


@pytest.fixture(name="app_context")
def fixture_app_context(app):
    """
    Yields the app context so that tests using the db will work.
    """
    with app.app_context() as c:
        yield c


@pytest.fixture(name="empty_db")
def fixture_empty_db(app_context):
    """
    Wipe the db.
    """
    lute.db.management.delete_all_data()


@pytest.fixture(name="client")
def fixture_demo_client(app):
    """
    Client using demo-data-loaded application.
    """
    return app.test_client()


def _get_test_language(lang_name):
    """
    Return language from the db if it already exists,
    or create it from the file.
    """
    lang = db.session.query(Language).filter(Language.name == lang_name).first()
    if lang is not None:
        return lang
    lang = get_language_def(lang_name)["language"]
    return lang


@pytest.fixture(name="spanish")
def fixture_spanish(app_context):
    return _get_test_language("Spanish")


@pytest.fixture(name="english")
def fixture_english(app_context):
    return _get_test_language("English")


@pytest.fixture(name="japanese")
def fixture_japanese(app_context):
    return _get_test_language("Japanese")


@pytest.fixture(name="turkish")
def fixture_turkish(app_context):
    return _get_test_language("Turkish")


@pytest.fixture(name="classical_chinese")
def fixture_cl_chinese(app_context):
    return _get_test_language("Classical Chinese")


@pytest.fixture(name="german")
def fixture_german(app_context):
    return _get_test_language("German")


@pytest.fixture(name="hindi")
def fixture_hindi(app_context):
    return _get_test_language("Hindi")



"""
Flask-wtf forms.
"""

from flask_wtf import FlaskForm
from wtforms import (
    StringField,
    IntegerField,
    BooleanField,
    SelectField,
    FormField,
    FieldList,
    Form,
    ValidationError,
)
from wtforms.validators import DataRequired
from lute.models.language import LanguageDictionary


class LanguageDictionaryForm(Form):
    """
    Language dictionary form, nested in Language form.
    """

    usefor = SelectField(
        choices=[("terms", "Terms"), ("sentences", "Sentences")],
        render_kw={"title": "Use dictionary for"},
    )
    dicttype = SelectField(
        choices=[
            ("embeddedhtml", "Embedded"),
            ("popuphtml", "Pop-up window"),
        ],
        render_kw={"title": "Show as"},
    )
    dicturi = StringField("URL", validators=[DataRequired()])
    is_active = BooleanField("Is active", render_kw={"title": "Is active?"})
    sort_order = IntegerField("Sort", render_kw={"style": "display: none"})

    def validate_dicturi(self, field):  # pylint: disable=unused-argument
        "Language must be set."
        # TODO TERM_for_dict_lookup_key: re-add <TERM> criteria.
        #  and "<TERM>" not in field.data:
        if "###" not in field.data:
            raise ValidationError("Dictionary URI must contain ###")  # or <TERM>


class LanguageForm(FlaskForm):
    """
    Language.
    """

    name = StringField("Name", validators=[DataRequired()])
    dictionaries = FieldList(
        FormField(LanguageDictionaryForm, default=LanguageDictionary)
    )
    show_romanization = BooleanField("Show Pronunciation field")
    right_to_left = BooleanField("Right-to-left")

    # Note!  The choices have to be set in the routes!
    # I originally had "choices=lute.parse.registry.supported_parsers()",
    # but it never worked: the Japanese mecab parser was excluded.
    # Possible coder error, not sure, but setting the choices at
    # form creation time works.
    parser_type = SelectField("Parse as", choices=[("tbd", "tbd")])

    character_substitutions = StringField("Character substitutions")

    regexp_split_sentences = StringField(
        "Split sentences at", validators=[DataRequired()]
    )
    exceptions_split_sentences = StringField("Split sentence exceptions")
    word_characters = StringField("Word characters", validators=[DataRequired()])

    def validate_dictionaries(self, field):  # pylint: disable=unused-argument
        "Dictionaries must be valid."

        # raise ValueError(self.dictionaries.data) # debugging
        def _get_actives(usefor):
            "Return dictionaries."
            return [
                d
                for d in self.dictionaries.data
                if d.get("usefor", "") == usefor and d.get("is_active")
            ]

        term_dicts = _get_actives("terms")
        sentence_dicts = _get_actives("sentences")
        if len(term_dicts) == 0:
            raise ValidationError("Please add an active Terms dictionary")
        if len(sentence_dicts) == 0:
            raise ValidationError("Please add an active Sentences dictionary")



"""
JapaneseParser tests.
"""

from lute.parse.base import ParsedToken


def assert_tokens_equals(text, lang, expected):
    """
    Parsing a text using a language should give the expected parsed tokens.

    expected is given as array of:
    [ original_text, is_word, is_end_of_sentence ]
    """
    p = lang.parser
    print("passing text:")
    print(text)
    actual = p.get_parsed_tokens(text, lang)
    expected = [ParsedToken(*a) for a in expected]
    assert [str(a) for a in actual] == [str(e) for e in expected]


def test_sample_1(classical_chinese):
    "Sample text parsed."
    s = "學而時習之，不亦說乎？"

    expected = [
        ["學", True],
        ["而", True],
        ["時", True],
        ["習", True],
        ["之", True],
        ["，", False],
        ["不", True],
        ["亦", True],
        ["說", True],
        ["乎", True],
        ["？", False, True],
    ]
    assert_tokens_equals(s, classical_chinese, expected)


def test_sample_2(classical_chinese):
    "Sample text parsed, spaces removed."
    s = """學  而時習 之，不亦說乎？
有朋    自遠方來，不亦樂乎？"""

    expected = [
        ["學", True],
        ["而", True],
        ["時", True],
        ["習", True],
        ["之", True],
        ["，", False],
        ["不", True],
        ["亦", True],
        ["說", True],
        ["乎", True],
        ["？", False, True],
        ["¶", False, True],
        ["有", True],
        ["朋", True],
        ["自", True],
        ["遠", True],
        ["方", True],
        ["來", True],
        ["，", False],
        ["不", True],
        ["亦", True],
        ["樂", True],
        ["乎", True],
        ["？", False, True],
    ]
    assert_tokens_equals(s, classical_chinese, expected)



"""
Term popup data tests.
"""

from lute.models.term import Term
from lute.read.service import get_popup_data
from lute.db import db


def test_term_with_no_parents(spanish, app_context):
    "Keep the lights on test, smoke only."
    t = Term(spanish, "gato")
    db.session.add(t)
    db.session.commit()

    d = get_popup_data(t.id)
    assert d["parentdata"] == [], "no parents"
    assert d["parentterms"] == "", "no parents"


def test_single_parent_data_not_added_if_same_translation(spanish, app_context):
    "Extra data added for display."
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "cat"
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = get_popup_data(t.id)
    assert d["parentdata"] == [], "no extra parent data"
    assert d["parentterms"] == "perro", "one parent"


def test_parent_data_added_if_parent_translation_is_different(spanish, app_context):
    "Extra data added for display."
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "kitty"
    t.parents.append(p)
    db.session.add(t)
    db.session.commit()

    d = get_popup_data(t.id)
    expected_p_data = {
        "term": "perro",
        "roman": None,
        "trans": "kitty",
        "tags": [],
    }
    assert d["parentdata"] == [expected_p_data], "extra parent data added"
    assert d["parentterms"] == "perro", "one parent"


def test_parent_data_always_added_if_multiple_parents(spanish, app_context):
    "Extra data added for display."
    t = Term(spanish, "gato")
    t.translation = "cat"
    p = Term(spanish, "perro")
    p.translation = "kitty"
    p2 = Term(spanish, "hombre")
    p2.translation = "kitteh"
    t.parents.append(p)
    t.parents.append(p2)
    db.session.add(t)
    db.session.commit()

    d = get_popup_data(t.id)
    expected_p1_data = {"term": "perro", "roman": None, "trans": "kitty", "tags": []}
    expected_p2_data = {"term": "hombre", "roman": None, "trans": "kitteh", "tags": []}
    expected = [expected_p1_data, expected_p2_data]
    assert d["parentdata"] == expected, "extra parent data added"
    assert d["parentterms"] == "perro, hombre", "parents"


def test_single_term_not_included_in_own_components(spanish, app_context):
    "Keep the lights on test, smoke only."
    t = Term(spanish, "gato")
    db.session.add(t)
    db.session.commit()

    d = get_popup_data(t.id)
    assert d["components"] == [], "no components"


def test_component_without_translation_not_returned(spanish, app_context):
    "Component word is returned."
    t = Term(spanish, "un gato")
    c = Term(spanish, "gato")
    db.session.add(t)
    db.session.add(c)
    db.session.commit()

    d = get_popup_data(t.id)
    assert d["components"] == [], "no data"


def test_component_word_with_translation_returned(spanish, app_context):
    "Component word is returned."
    t = Term(spanish, "un gato")
    c = Term(spanish, "gato")
    c.translation = "cat"
    db.session.add(t)
    db.session.add(c)
    db.session.commit()

    d = get_popup_data(t.id)
    c_data = {"term": "gato", "roman": None, "trans": "cat", "tags": []}
    assert d["components"] == [c_data], "one component"


def test_nested_multiword_components(spanish, app_context):
    "Complete components are returned."
    t = Term(spanish, "un gato gordo")
    db.session.add(t)
    for c in [("gato", "cat"), ("gat", "x"), ("gato gordo", ""), ("un gato", "a cat")]:
        ct = Term(spanish, c[0])
        ct.translation = c[1]
        db.session.add(ct)
    db.session.commit()

    d = get_popup_data(t.id)
    c_data = [
        {"term": "gato", "roman": None, "trans": "cat", "tags": []},
        {"term": "un\u200b \u200bgato", "roman": None, "trans": "a cat", "tags": []},
    ]
    assert d["components"] == c_data, "components"



"""
Helper methods to get data for datatables display.
"""

import re
from sqlalchemy.sql import text
from lute.parse.registry import supported_parser_types


def supported_parser_type_criteria():
    "Helper to get all supported parser_types."
    typecrit = [f"'{p}'" for p in supported_parser_types()]
    typecrit.append("'zz_dummy_parser'")
    return ",".join(typecrit)


class DataTablesFlaskParamParser:
    """
    Parse datatables form parameters into the structure needed for
    DataTablesSqliteQuery.

    The standard datatables ajax post gives parameters like the following:

    draw: 1
    columns[0][data]: 0
    columns[0][name]: BkTitle
    columns[0][searchable]: true
    columns[0][orderable]: true

    Flask converts that to this, more or less:

    {
      draw: 1
      columns[0][data]: 0
      columns[0][name]: BkTitle
      columns[0][searchable]: true
      columns[0][orderable]: true
    }

    But the query helper requires parameters like this:

    {
      draw: 1,
      columns: [
        { data: 0, name: BkTitle, ... }
    }

    All code here adapted from https://github.com/coding-doc/
    sqlalchemy2-datatables/blob/main/src/datatables/datatable.py.
    """

    @staticmethod
    def _parse_order(request_params):
        """Parse the order[index][*] parameters."""
        order = []
        order_re = re.compile(r"order\[(.*?)]\[column]")
        order_params = {k: v for k, v in request_params.items() if order_re.match(k)}

        for i in range(len(order_params)):
            dt_column_order = {
                "column": int(request_params[f"order[{i}][column]"]),
                "dir": request_params[f"order[{i}][dir]"],
            }
            order.append(dt_column_order)
        return order

    @staticmethod
    def _parse_columns(request_params):
        """Parse the column[index][*] parameters."""
        columns = []
        # Extract only the keys of type columns[i][data] from the params
        data_re = re.compile(r"columns\[(.*?)]\[data]")
        data_param = {k: v for k, v in request_params.items() if data_re.match(k)}

        for i in range(len(data_param)):
            column = {
                "index": i,
                "data": data_param.get(f"columns[{i}][data]"),
                "name": request_params.get(f"columns[{i}][name]"),
                "searchable": request_params.get(f"columns[{i}][searchable]") == "true",
                "orderable": request_params.get(f"columns[{i}][orderable]") == "true",
                "search": {
                    "value": request_params.get(f"columns[{i}][search][value]"),
                    "regex": request_params.get(f"columns[{i}][search][regex]")
                    == "true",
                },
            }
            columns.append(column)
        return columns

    @staticmethod
    def parse_params(requestform) -> dict:
        """Parse the request (query) parameters."""
        request_params = requestform.to_dict(flat=True)

        return {
            "draw": int(request_params.get("draw", 1)),
            "start": int(request_params.get("start", 0)),
            "length": int(request_params.get("length", -1)),
            "search": {
                "value": request_params.get("search[value]"),
                "regex": request_params.get("search[regex]") == "true",
            },
            "columns": DataTablesFlaskParamParser._parse_columns(request_params),
            "order": DataTablesFlaskParamParser._parse_order(request_params),
        }


class DataTablesSqliteQuery:
    "Get data for datatables rendering."

    @staticmethod
    def where_and_params(searchable_cols, parameters):
        "Build where string and get the 'where' parameters."

        search = parameters["search"]
        search_string = search["value"] if search["value"] is not None else ""
        search_string = search_string.strip()

        search_parts = search_string.split()
        search_parts = list(filter(lambda p: len(p) > 0, search_parts))

        # If no searchable columns or search params, stop.
        if len(searchable_cols) == 0 or len(search_parts) == 0:
            return ["", {}]

        params = {}
        part_wheres = []
        for i, p in enumerate(search_parts):
            lwild = "%" if not p.startswith("^") else ""
            rwild = "%" if not p.endswith("$") else ""
            p = p.lstrip("^").rstrip("$")
            params[f"s{i}"] = p

            col_wheres = []
            for cname in searchable_cols:
                col_wheres.append(f"{cname} LIKE '{lwild}' || :s{i} || '{rwild}'")

            part_wheres.append("(" + " OR ".join(col_wheres) + ")")

        return ["WHERE " + " AND ".join(part_wheres), params]

    @staticmethod
    def get_sql(base_sql, parameters):
        "Build sql used for datatables queries."
        columns = parameters["columns"]

        def cols_with(attr):
            return [c["name"] for c in columns if c[attr] is True]

        orderby = ", ".join(cols_with("orderable"))
        for order in parameters["order"]:
            col_index = int(order["column"])
            sort_field = columns[col_index]["name"]
            orderby = f"{sort_field} {order['dir']}, {orderby}"
        orderby = f"ORDER BY {orderby}"

        [where, params] = DataTablesSqliteQuery.where_and_params(
            cols_with("searchable"), parameters
        )

        realbase = f"({base_sql}) realbase".replace("\n", " ")
        select_field_list = ", ".join([c["name"] for c in columns if c["name"] != ""])

        start = parameters["start"]
        length = parameters["length"]
        # pylint: disable=line-too-long
        data_sql = f"SELECT {select_field_list} FROM (select * from {realbase} {where} {orderby} LIMIT {start}, {length}) src {orderby}"

        return {
            "recordsTotal": f"select count(*) from {realbase}",
            "recordsFiltered": f"select count(*) from {realbase} {where}",
            "data": data_sql,
            "params": params,
            "draw": int(parameters["draw"]),
        }

    @staticmethod
    def get_data(base_sql, parameters, conn):
        "Return dict required for datatables rendering."
        recordsTotal = None
        recordsFiltered = None

        try:
            sql_dict = DataTablesSqliteQuery.get_sql(base_sql, parameters)

            def runqry(name, use_params=True):
                "Run the given query from the datatables list of queries."
                prms = None
                if use_params:
                    prms = sql_dict["params"]
                return conn.execute(text(sql_dict[name]), prms)

            recordsTotal = runqry("recordsTotal", False).fetchone()[0]
            recordsFiltered = runqry("recordsFiltered").fetchone()[0]
            res = runqry("data")
            ret = [list(row) for row in res.fetchall()]
        except Exception as e:
            raise e

        result = {
            "recordsTotal": recordsTotal,
            "recordsFiltered": recordsFiltered,
            "data": ret,
        }
        return result



"""
Book tests.
"""

from lute.termtag.datatables import get_data_tables_list


def test_smoke_datatables_query_runs(app_context):
    """
    Smoke test only, ensure query runs.
    """
    columns = [
        {"data": "0", "name": "TgID", "searchable": False, "orderable": False},
        {"data": "1", "name": "TgText", "searchable": True, "orderable": True},
    ]
    params = {
        "draw": "1",
        "columns": columns,
        "order": [{"column": "1", "dir": "asc"}],
        "start": "10",
        "length": "50",
        "search": {"value": "", "regex": False},
    }

    d = get_data_tables_list(params)
    print(d)
    a = 1
    assert a == 1, "dummy check"



"""
Factory.

Methods: create_app.
"""

import os
import json
import platform
import traceback
from flask import (
    Flask,
    render_template,
    request,
    redirect,
    flash,
    current_app,
    make_response,
    send_from_directory,
    jsonify,
)
from sqlalchemy.event import listens_for
from sqlalchemy.pool import Pool

from lute.config.app_config import AppConfig
from lute.db import db
from lute.db.setup.main import setup_db
from lute.db.data_cleanup import clean_data
import lute.backup.service as backupservice
import lute.db.demo
import lute.utils.formutils

from lute.models.book import Book
from lute.models.language import Language
from lute.models.setting import BackupSettings, UserSetting
from lute.book.stats import refresh_stats, mark_stale

from lute.book.routes import bp as book_bp
from lute.language.routes import bp as language_bp
from lute.term.routes import bp as term_bp
from lute.termtag.routes import bp as termtag_bp
from lute.read.routes import bp as read_bp
from lute.bing.routes import bp as bing_bp
from lute.userimage.routes import bp as userimage_bp
from lute.useraudio.routes import bp as useraudio_bp
from lute.termimport.routes import bp as termimport_bp
from lute.backup.routes import bp as backup_bp
from lute.dev_api.routes import bp as dev_api_bp
from lute.settings.routes import bp as settings_bp
from lute.themes.routes import bp as themes_bp
from lute.stats.routes import bp as stats_bp
from lute.cli.commands import bp as cli_bp


def _setup_app_dirs(app_config):
    """
    App needs the data dir, backups, and other directories.
    """
    dp = app_config.datapath
    required_dirs = [
        {"d": dp, "readme": "Lute data folder."},
        {
            "d": app_config.default_user_backup_path,
            "readme": "Default path for user backups, can be overridden in settings.",
        },
        {
            "d": app_config.system_backup_path,
            "readme": "Database backups created by Lute at app start, just in case.",
        },
        {
            "d": app_config.userimagespath,
            "readme": "User images.  Each subfolder is a language's ID.",
        },
        {
            "d": app_config.useraudiopath,
            "readme": "User audio.  Each file is a book's audio.",
        },
        {
            "d": app_config.temppath,
            "readme": "Temp directory for export file writes, to avoid permissions issues.",
        },
    ]
    for rec in required_dirs:
        d = rec["d"]
        if not os.path.exists(d):
            os.makedirs(d)
        readme = os.path.join(d, "README.md")
        if not os.path.exists(readme):
            with open(readme, "w", encoding="utf-8") as f:
                f.write(rec["readme"])


def _add_base_routes(app, app_config):
    """
    Add some basic routes.
    """

    @app.context_processor
    def inject_menu_bar_vars():
        """
        Inject backup settings into the all templates for the menu bar.
        """
        bs = BackupSettings.get_backup_settings()
        have_languages = len(db.session.query(Language).all()) > 0
        ret = {
            "have_languages": have_languages,
            "backup_enabled": bs.backup_enabled,
            "backup_directory": bs.backup_dir,
            "backup_last_display_date": bs.last_backup_display_date,
            "backup_time_since": bs.time_since_last_backup,
            "user_settings": json.dumps(UserSetting.all_settings()),
        }
        return ret

    @app.route("/")
    def index():
        is_production = not lute.db.demo.contains_demo_data()
        bkp_settings = BackupSettings.get_backup_settings()

        have_books = len(db.session.query(Book).all()) > 0
        have_languages = len(db.session.query(Language).all()) > 0
        language_choices = lute.utils.formutils.language_choices("(all languages)")
        current_language_id = lute.utils.formutils.valid_current_language_id()

        should_run_auto_backup = backupservice.should_run_auto_backup(bkp_settings)
        # Only back up if we have books, otherwise the backup is
        # kicked off when the user empties the demo database.
        if is_production and have_books and should_run_auto_backup:
            return redirect("/backup/backup", 302)

        refresh_stats()
        warning_msg = backupservice.backup_warning(bkp_settings)
        backup_show_warning = (
            bkp_settings.backup_warn
            and bkp_settings.backup_enabled
            and warning_msg != ""
        )

        return render_template(
            "index.html",
            hide_homelink=True,
            dbname=app_config.dbname,
            datapath=app_config.datapath,
            tutorial_book_id=lute.db.demo.tutorial_book_id(),
            have_books=have_books,
            have_languages=have_languages,
            language_choices=language_choices,
            current_language_id=current_language_id,
            is_production_data=is_production,
            # Backup stats
            backup_show_warning=backup_show_warning,
            backup_warning_msg=warning_msg,
        )

    @app.route("/refresh_all_stats")
    def refresh_all_stats():
        books_to_update = db.session.query(Book).filter(Book.archived == 0).all()

        for book in books_to_update:
            mark_stale(book)

        refresh_stats()

        return redirect("/", 302)

    @app.route("/wipe_database")
    def wipe_db():
        if lute.db.demo.contains_demo_data():
            lute.db.demo.delete_demo_data()
            msg = """
            The database has been wiped clean.  Have fun! <br /><br />
            <i>(Lute has automatically enabled backups --
            change your <a href="/settings/index">Settings</a> as needed.)</i>
            """
            flash(msg)
        return redirect("/", 302)

    @app.route("/remove_demo_flag")
    def remove_demo():
        if lute.db.demo.contains_demo_data():
            lute.db.demo.remove_flag()
            msg = """
            Demo mode deactivated. Have fun! <br /><br />
            <i>(Lute has automatically enabled backups --
            change your <a href="/settings/index">Settings</a> as needed.)</i>
                        """
            flash(msg)
        return redirect("/", 302)

    @app.route("/version")
    def show_version():
        ac = current_app.env_config
        return render_template(
            "version.html",
            version=lute.__version__,
            datapath=ac.datapath,
            database=ac.dbfilename,
            is_docker=ac.is_docker,
        )

    @app.route("/hotkeys")
    def show_hotkeys():
        return render_template("hotkeys.html")

    @app.route("/info")
    def show_info():
        """
        Json return of some data.

        Used in lute.verify module for tests.

        This likely belongs in a different 'api' location,
        but leaving it here for now.
        """
        ret = {
            "version": lute.__version__,
            "datapath": current_app.config["DATAPATH"],
            "database": current_app.config["DATABASE"],
        }
        return jsonify(ret)

    @app.route("/static/js/never_cache/<path:filename>")
    def custom_js(filename):
        """
        Some files should never be cached.
        """
        response = make_response(send_from_directory("static/js", filename))
        response.headers[
            "Cache-Control"
        ] = "no-store, no-cache, must-revalidate, max-age=0"
        return response

    @app.errorhandler(500)
    def _internal_server_error(e):  # pylint: disable=unused-argument
        """
        Custom error handler for 500 Internal Server Error
        """
        exception_info = traceback.format_exc()
        # Should add logging ...
        # app.logger.error(exception_info)
        return (
            render_template(
                "errors/500_error.html",
                exception_info=exception_info,
                version=lute.__version__,
                platform=platform.platform(),
                is_docker=current_app.env_config.is_docker,
            ),
            500,
        )

    @app.errorhandler(404)
    def _page_not_found(e):  # pylint: disable=unused-argument
        "Show custom error page on 404."
        return (
            render_template(
                "errors/404_error.html",
                version=lute.__version__,
                requested_url=request.url,
                referring_page=request.referrer,
            ),
            404,
        )


def _create_app(app_config, extra_config):
    """
    Create the app using the given configuration,
    and init the SqlAlchemy db.
    """

    app = Flask(__name__, instance_path=app_config.datapath)

    config = {
        "SECRET_KEY": "some_secret",
        "DATABASE": app_config.dbfilename,
        "ENV": app_config.env,
        "SQLALCHEMY_DATABASE_URI": f"sqlite:///{app_config.dbfilename}",
        "DATAPATH": app_config.datapath,
        # ref https://flask-sqlalchemy.palletsprojects.com/en/2.x/config/
        # Don't track mods.
        "SQLALCHEMY_TRACK_MODIFICATIONS": False,
    }

    final_config = {**config, **extra_config}
    app.config.from_mapping(final_config)

    # Attach the app_config to app so it's available at runtime.
    app.env_config = app_config

    db.init_app(app)

    @listens_for(Pool, "connect")
    def _pragmas_on_connect(dbapi_con, con_record):  # pylint: disable=unused-argument
        dbapi_con.execute("pragma recursive_triggers = on;")

    with app.app_context():
        db.create_all()
        UserSetting.load()
        # TODO valid parsers: do parser check, mark valid as active, invalid as inactive.
        clean_data()
    app.db = db

    _add_base_routes(app, app_config)
    app.register_blueprint(language_bp)
    app.register_blueprint(book_bp)
    app.register_blueprint(term_bp)
    app.register_blueprint(termtag_bp)
    app.register_blueprint(read_bp)
    app.register_blueprint(bing_bp)
    app.register_blueprint(userimage_bp)
    app.register_blueprint(useraudio_bp)
    app.register_blueprint(termimport_bp)
    app.register_blueprint(backup_bp)
    app.register_blueprint(settings_bp)
    app.register_blueprint(themes_bp)
    app.register_blueprint(stats_bp)
    app.register_blueprint(cli_bp)
    if app_config.is_test_db:
        app.register_blueprint(dev_api_bp)

    return app


def create_app(
    app_config_path=None,
    extra_config=None,
    output_func=None,
):
    """
    App factory.  Calls dbsetup, and returns Flask app.

    Args:
    - app_config_path: path to yml file.  If None, use root config or default.
    - extra_config: dict, e.g. pass { 'TESTING': True } during unit tests.
    """

    if app_config_path is None:
        if os.path.exists("config.yml"):
            app_config_path = "config.yml"
        else:
            app_config_path = AppConfig.default_config_filename()

    app_config = AppConfig(app_config_path)
    _setup_app_dirs(app_config)
    setup_db(app_config, output_func)

    if extra_config is None:
        extra_config = {}
    app = _create_app(app_config, extra_config)

    return app



"""
User images routes.

User images are stored in the database as /userimages/langid/term, but
with no jpeg extension.  Reason: the old symfony code couldn't manage
urls with periods.
"""

import os
from flask import Blueprint, send_from_directory, current_app

bp = Blueprint("userimages", __name__, url_prefix="/userimages")


@bp.route("/<int:lgid>/<term>", methods=["GET"])
def get_image(lgid, term):
    "Serve the image from the data/userimages directory."
    datapath = current_app.config["DATAPATH"]
    directory = os.path.join(datapath, "userimages", str(lgid))
    filename = term + ".jpeg"
    return send_from_directory(directory, filename)



"""
Calculating what items should be rendered in the browser.
"""

import re
import functools
from lute.models.language import Language
from lute.models.term import Term, Status


class RenderableCalculator:
    """
    Calculating what TextTokens and Terms should be rendered.

    Suppose we had the following TextTokens A-I, with spaces between:

     A B C D E F G H I

    Then suppose we had the following Terms:
      "B C"       (term J)
      "E F G H I" (K)
      "F G"       (L)
      "C D E"     (M)

    Stacking these:

     A B C D E F G H I

      "B C"              (term J)
            "E F G H I"  (term K)
              "F G"      (term L)
        "C D E"          (term M)

    We can say:

    - term J "contains" TextTokens B and C, so B and C should not be rendered.
    - K contains tokens E-I, and also L, so none of those should be rendered.
    - M is _not_ contained by anything else, so it should be rendered.
    """

    def _tokens_to_string(self, texttokens):
        "Convert array of tokens to string for debugging."
        arr = [f"{tok.order}: '{tok.token}'" for tok in texttokens]
        return "\n".join(arr)

    def _assert_texttokens_are_contiguous(self, texttokens):
        "Check ordering."
        prevtok = None
        for tok in texttokens:
            if prevtok is not None and prevtok.order != (tok.order - 1):
                toks = self._tokens_to_string(texttokens)
                msg = f"Order error at pos {prevtok.order}:\n{toks}"
                raise RuntimeError(msg)
            prevtok = tok

    def _get_renderable(
        self, tokenlocator, terms, texttokens
    ):  # pylint: disable=too-many-locals
        """
        Return RenderableCandidates that will **actually be rendered**.

        Method to determine what should be rendered:

        1. Create a "rendered array".  On completion of this algorithm,
        each position in the array will be filled with the ID of the
        RenderableCandidate that should actually appear there (and
        which might hide other candidates).

        2. Start by saying that all the original texttokens will be
        rendered by writing each candidate ID in the rendered array.

        3. Create candidates for all the terms.

        4. Starting with the shortest terms first (fewest text tokens),
        and starting _at the end_ of the string, "write" the candidate
        ID to the output "rendered array", for each token in the candidate.

        At the end of this process, each position in the "rendered array"
        should be filled with the ID of the corresponding candidate
        that will actually appear in that position.  By getting the
        unique IDs and returning just their candidates, we should have
        the list of candidates that would be "visible" on render.

        Applying the above algorithm to the example given in the class
        header:

        We have the following TextTokens A-I, with spaces between:

         a b c d e f g h i

        And the following terms, arranged from shortest to longest:
          "B C"
          "F G"
          "C D E"
          "E F G H I"

        First, terms are created for each individual token in the
        original string:

        A B C D E F G H I

        Then the positions for each of the terms are calculated:

        [A B C D E F G H I]

          "B C"
                  "F G"
            "C D E"
                "E F G H I"

        Then, "writing" terms order by their length, and then by their
        distance from the *end* of the string:

        - "F G" is written first, because it's short, and is nearest
          the end:
          => "A B C D E [F-G] H I"
        - "B C" is next:
          => "A [B-C] D E [F-G] H I"
        - then "C D E":
          => "A [B-C][C-D-E] [F-G] H I"
        then "E F G H I":
          => "A [B-C][C-D-E][E-F-G-H-I]"
        """

        # All the candidates to be considered for rendering.
        candidates = {}

        # Step 1.  Map of the token position to the id of the
        # candidate that should be rendered there.
        rendered = {}

        # Step 2 - fill with the original texttokens.
        for tok in texttokens:
            rc = RenderableCandidate()
            rc.display_text = tok.token
            rc.text = tok.token
            rc.pos = tok.order
            rc.is_word = tok.is_word
            candidates[rc.id] = rc
            rendered[rc.pos] = rc.id

        # 3.  Create candidates for all the terms.
        termcandidates = []

        foundterms = [t for t in terms if t.text_lc in tokenlocator.subjLC]
        for term in foundterms:
            for loc in tokenlocator.locate_string(term.text_lc):
                rc = RenderableCandidate()
                rc.term = term
                rc.display_text = loc["text"]
                rc.text = loc["text"]
                rc.pos = texttokens[0].order + loc["index"]
                rc.length = term.token_count
                rc.is_word = 1

                termcandidates.append(rc)
                candidates[rc.id] = rc

        # 4a.  Sort the term candidates: first by length, then by position.
        def compare(a, b):
            # Longest sorts first.
            if a.length != b.length:
                return -1 if (a.length > b.length) else 1
            # Lowest position (closest to front of string) sorts first.
            return -1 if (a.pos < b.pos) else 1

        termcandidates.sort(key=functools.cmp_to_key(compare))

        # The termcandidates should now be sorted such that longest
        # are first, with items of equal length being sorted by
        # position.  By traversing this in reverse and "writing"
        # their IDs to the "rendered" array, we should end up with
        # the final IDs in each position.
        termcandidates.reverse()
        for tc in termcandidates:
            for i in range(tc.length):
                rendered[tc.pos + i] = tc.id

        rcids = list(set(rendered.values()))
        return [candidates[rcid] for rcid in rcids]

    def _sort_by_order_and_tokencount(self, items):
        items.sort(key=lambda x: (x.pos, -x.length))
        return items

    def _calc_overlaps(self, items):
        for i in range(1, len(items)):
            prev = items[i - 1]
            curr = items[i]

            prevterm_last_token_pos = prev.pos + prev.length - 1
            overlap = prevterm_last_token_pos - curr.pos + 1

            if overlap > 0:
                zws = chr(0x200B)
                curr_tokens = curr.text.split(zws)
                show = curr_tokens[overlap:]
                curr.display_text = zws.join(show)

        return items

    def main(self, language, words, texttokens):
        """
        Main entrypoint.

        Given a language and some terms and texttokens,
        return the RenderableCandidates to be rendered.
        """
        texttokens.sort(key=lambda x: x.order)
        self._assert_texttokens_are_contiguous(texttokens)

        subject = TokenLocator.make_string([t.token for t in texttokens])
        tocloc = TokenLocator(language, subject)

        renderable = self._get_renderable(tocloc, words, texttokens)
        items = self._sort_by_order_and_tokencount(renderable)
        items = self._calc_overlaps(items)
        return items

    @staticmethod
    def get_renderable(lang, words, texttokens):
        "Convenience method, calls main."
        rc = RenderableCalculator()
        return rc.main(lang, words, texttokens)


class RenderableCandidate:  # pylint: disable=too-many-instance-attributes
    """
    An item that may or may not be rendered on the browser.

    Given some Terms contained in a text, the RenderableCalculator
    creates RenderableCandidates for each Term in the text, as well as
    the original text tokens.

    When the final set of actually rendered RenderableCandidates is
    found (with self.render is True), these are convered into TextItems
    for the final render.
    """

    # ID incremented for each instance.
    class_id = 0

    def __init__(self):
        RenderableCandidate.class_id += 1

        self.id: int = RenderableCandidate.class_id
        self.term: Term = None
        self.display_text: str = None  # Text to show, if there is any overlap
        self.text: str = None  # Actual text of the term
        self.text_lc: str = None
        self.pos: int = None
        self.length: int = 1
        self.is_word: int = None
        self.render: bool = True

    def __repr__(self):
        parts = [f"pos {self.pos}", f"render {self.render}" f"(id {self.id})"]
        parts = " ".join(parts)
        return f'<RenderableCandidate "{self.text}", {parts}>'

    @property
    def term_id(self) -> int:
        return self.term.id if self.term else None

    @property
    def order_end(self) -> int:
        return self.pos + self.length - 1

    def make_text_item(self, p_num: int, se_id: int, lang: Language):
        """
        Create a TextItem for final rendering.
        """
        t = TextItem(self.term)
        t.order = self.pos
        t.lang_id = lang.id
        t.display_text = self.display_text
        t.text = self.text
        t.text_lc = lang.get_lowercase(self.text)
        t.token_count = self.length
        t.para_id = p_num
        t.se_id = se_id
        t.is_word = self.is_word
        t.text_length = len(self.text)
        return t


class TokenLocator:
    """
    Helper class for finding tokens and positions in a subject string.

    Finds a given token (word) in a sentence, ignoring case, returning
    the actual word in the sentence (its original case), and its index
    or indices.

    For example, given:

      - $subject "/this/ /CAT/ /is/ /big/"
      - $find_patt = "/cat/"

    (where "/" is the zero-width space to indicate word boundaries)

    this method would return [ { 'term': "CAT", 'index': 2 } ]
      - the token "cat" is actually "CAT" (uppercase) in the sentence
      - it's at index = 2

    Note that the language of the string must also be provided, because
    some languages (Turkish!) have unusual case requirements.

    See the test cases for more examples.
    """

    def __init__(self, language, subject):
        self.language = language
        self.subject = subject
        self.subjLC = self.language.get_lowercase(self.subject)

    def locate_string(self, s):
        """
        Find the string s in the subject self.subject.
        """
        find_lc = self.language.get_lowercase(s)
        find_lc = TokenLocator.make_string(find_lc)

        matches = self.preg_match_capture(find_lc, self.subjLC)

        # The matches were performed with the lowercased subject,
        # because some languages (Turkish!) have funny cases.
        # We need to convert the matched text back to the _original_
        # subject string cases.
        subj = self.subject

        def make_text_index_pair(match):
            matchtext = match[0]  # includes zws at start and end.
            matchlen = len(matchtext)
            matchpos = match[1]

            # print(f"found match \"{matchtext}\" len={matchlen} pos={matchpos}")
            original_subject_text = subj[matchpos : matchpos + matchlen]
            zws = "\u200B"
            t = original_subject_text.lstrip(zws).rstrip(zws)
            index = self.get_count_before(subj, matchpos)
            return {"text": t, "index": index}

        termmatches = list(map(make_text_index_pair, matches))

        return termmatches

    def get_count_before(self, string, pos):
        """
        Count of tokens found in string before position pos.
        """
        zws = "\u200B"
        beforesubstr = string[:pos]
        n = beforesubstr.count(zws)
        return n

    def preg_match_capture(self, find_lc, subject):
        """
        Return the matched text and their start positions in the subject.

        E.g. search for r'cat' in "there is a CAT and a Cat" returns:
        [['CAT', 11], ['Cat', 21]]
        """

        # "(?=())" is required because sometimes the search pattern can
        # overlap -- e.g. _b_b_ has _b_ *twice*.
        # https://stackoverflow.com/questions/5616822/
        #   how-to-use-regex-to-find-all-overlapping-matches
        pattern = rf"(?=({re.escape(find_lc)}))"

        matches = re.finditer(pattern, subject, flags=re.IGNORECASE)

        # Use group(1) to get the match text, because group(0) is a
        # zero-length string.
        result = [[match.group(1), match.start()] for match in matches]
        return result

    @staticmethod
    def make_string(t):
        """
        Append zero-width string to string to simplify/standardize searches.
        """
        zws = "\u200B"
        if isinstance(t, list):
            t = zws.join(t)
        return f"{zws}{t}{zws}"


class TextItem:  # pylint: disable=too-many-instance-attributes
    """
    Unit to be rendered.

    Data structure for template read/textitem.html

    Some elements are lazy loaded, because they're only needed in
    certain situations.
    """

    def __init__(self, term=None):
        self.lang_id: int
        self.order: int
        self.text: str  # The original, un-overlapped text.
        self.display_text: str  # The actual text to display on screen.
        # If part of the text has been overlapped by a
        # prior token, this will be different from Text.
        self.token_count: int
        self.text_lc: str
        self.para_id: int
        self.se_id: int
        self.is_word: int
        self.text_length: int

        # Calls setter
        self.term = term

        # The tooltip should be shown for well-known/ignored TextItems
        # that merit a tooltip. e.g., if there isn't any actual Term
        # entity associated with this TextItem, nothing more is needed.
        # Also, if there is a Term entity but it's mostly empty, a
        # tooltip isn't useful.
        self._show_tooltip: bool = None

        # The flash message can be None, so we need an extra flag
        # to determine if it has been loaded or not.
        self._flash_message_loaded: bool = False
        self._flash_message: str = None

    def __repr__(self):
        return f'<TextItem "{self.text}" (wo_id={self.wo_id})>'

    @property
    def term(self):
        return self._term

    @term.setter
    def term(self, t):
        self.wo_id = None
        self.wo_status = None
        self._term = t
        if t is None:
            return

        self.wo_id = t.id
        self.wo_status = t.status
        if t.status >= 1 and t.status <= 5:
            self._show_tooltip = True

    @property
    def show_tooltip(self):
        """
        Show the tooltip if there is anything to show.
        Lazy loaded as needed.
        """
        if self._show_tooltip is not None:
            return self._show_tooltip
        if self.term is None:
            return False

        def blank_string(s):
            return s is None or s.strip() == ""

        def has_extra(cterm):
            if cterm is None:
                return False
            no_extra = (
                blank_string(cterm.translation)
                and blank_string(cterm.romanization)
                and cterm.get_current_image() is None
            )
            return not no_extra

        self._show_tooltip = has_extra(self.term)
        for p in self.term.parents:
            self._show_tooltip = self._show_tooltip or has_extra(p)
        return self._show_tooltip

    @property
    def flash_message(self):
        """
        Return flash message if anything present.
        Lazy loaded as needed.
        """
        if self._flash_message_loaded:
            return self._flash_message
        if self.term is None:
            return None

        self._flash_message = self.term.get_flash_message()
        self._flash_message_loaded = True
        return self._flash_message

    @property
    def html_display_text(self):
        """
        Content to be rendered to browser.
        """
        zws = chr(0x200B)
        return self.display_text.replace(zws, "").replace(" ", "&nbsp;")

    @property
    def span_id(self):
        """
        Each span gets a unique id.
        Arbitrary format: ID-{order}-{tokencount}.

        This *might* not be necessary ... I don't think IDs are used anywhere.
        """
        parts = ["ID", str(self.order), str(max(1, self.token_count))]
        return "-".join(parts)

    @property
    def status_class(self):
        "Status class to apply."
        if self.wo_id is None:
            return "status0"
        return f"status{self.wo_status}"

    @property
    def html_class_string(self):
        """
        Create class string for TextItem.
        """
        if self.is_word == 0:
            return "textitem"

        if self.wo_id is None:
            classes = ["textitem", "click", "word"]
            return " ".join(classes)

        st = self.wo_status
        classes = [
            "textitem",
            "click",
            "word",
            "word" + str(self.wo_id),
        ]

        tooltip = (
            st not in (Status.WELLKNOWN, Status.IGNORED, Status.UNKNOWN)
            or self.show_tooltip
            or self.flash_message is not None
        )
        if tooltip:
            classes.append("showtooltip")

        if self.flash_message is not None:
            classes.append("hasflash")

        if self.display_text != self.text:
            classes.append("overlapped")

        return " ".join(classes)



"""
Book statistics.
"""

import json
from lute.read.render.service import get_paragraphs
from lute.db import db
from lute.models.book import Book


def _last_5_pages(book, txindex):
    "Get next 5 pages, or at least 5 pages."
    start_index = max(0, txindex - 5)
    end_index = txindex + 5
    texts = book.texts[start_index:end_index]
    return texts[-5:]


def get_status_distribution(book):
    """
    Return statuses and count of unique words per status.

    Does a full render of a small number of pages
    to calculate the distribution.
    """
    txindex = 0

    if (book.current_tx_id or 0) != 0:
        for t in book.texts:
            if t.id == book.current_tx_id:
                break
            txindex += 1

    # get next 5 pages, a good enough sample ...
    # min 5 pages.
    text_sample = [t.text for t in _last_5_pages(book, txindex)]
    text_sample = "\n".join(text_sample)

    paras = get_paragraphs(text_sample, book.language)

    def flatten_list(nested_list):
        result = []
        for item in nested_list:
            if isinstance(item, list):
                result.extend(flatten_list(item))
            else:
                result.append(item)
        return result

    text_items = []
    for s in flatten_list(paras):
        text_items.extend(s.textitems)
    text_items = [ti for ti in text_items if ti.is_word]

    statterms = {0: [], 1: [], 2: [], 3: [], 4: [], 5: [], 98: [], 99: []}

    for ti in text_items:
        statterms[ti.wo_status or 0].append(ti.text_lc)

    stats = {}
    for statusval, allterms in statterms.items():
        uniques = list(set(allterms))
        statterms[statusval] = uniques
        stats[statusval] = len(uniques)

    return stats


##################################################
# Stats table refresh.


class BookStats(db.Model):
    "The stats table."
    __tablename__ = "bookstats"

    id = db.Column(db.Integer, primary_key=True)
    BkID = db.Column(db.Integer)
    distinctterms = db.Column(db.Integer)
    distinctunknowns = db.Column(db.Integer)
    unknownpercent = db.Column(db.Integer)
    status_distribution = db.Column(db.String, nullable=True)


def refresh_stats():
    "Refresh stats for all books requiring update."
    books_to_update = (
        db.session.query(Book)
        .filter(~Book.id.in_(db.session.query(BookStats.BkID)))
        .all()
    )
    books = [b for b in books_to_update if b.is_supported]
    for book in books:
        stats = _get_stats(book)
        _update_stats(book, stats)


def mark_stale(book):
    "Mark a book's stats as stale to force refresh."
    bk_id = book.id
    db.session.query(BookStats).filter_by(BkID=bk_id).delete()
    db.session.commit()


def _get_stats(book):
    "Calc stats for the book using the status distribution."
    status_distribution = get_status_distribution(book)
    unknowns = status_distribution[0]
    allunique = sum(status_distribution.values())

    percent = 0
    if allunique > 0:  # In case not parsed.
        percent = round(100.0 * unknowns / allunique)

    sd = json.dumps(status_distribution)

    # Any change in the below fields requires a change to
    # update_stats as well, query insert doesn't check field order.
    return [allunique, unknowns, percent, sd]


def _update_stats(book, stats):
    "Update BookStats for the given book."
    new_stats = BookStats(
        BkID=book.id,
        distinctterms=stats[0],
        distinctunknowns=stats[1],
        unknownpercent=stats[2],
        status_distribution=stats[3],
    )
    db.session.add(new_stats)
    db.session.commit()



"""
Verification that Lute is running correctly.

Used in github actions to verify running app.

Usage:

python -m utils.verify [port]
"""

import sys
import json
import requests


def verify(port):
    """
    Check the /info page.
    """

    url = f"http://localhost:{port}/info"
    resp = requests.get(url, timeout=5)
    c = resp.status_code

    if c != 200:
        raise RuntimeError(f"Code {c} for url {url}")

    print("Lute is running:")
    print(json.dumps(resp.json(), indent=2))


if __name__ == "__main__":
    useport = None
    if len(sys.argv) == 2:
        useport = sys.argv[1]
    else:
        print("Must supply port as argument")
        sys.exit(1)
    verify(useport)



"""
Read service tests.
"""

# from lute.db import db
from lute.language import service
from tests.dbasserts import assert_sql_result


def test_get_all_lang_defs(app_context):
    "Can get all predefined languages."
    defs = service.get_defs()
    engs = [d for d in defs if d["language"].name == "English"]
    assert len(engs) == 1, "have english"
    eng = engs[0]
    assert len(eng["books"]) == 2, "tutorial and follow-up"
    titles = [b.title for b in eng["books"]]
    titles.sort()
    assert titles == ["Tutorial", "Tutorial follow-up"], "book titles"


def test_get_language_def():
    """
    Smoke test, can load a new language from yaml definition.
    """
    lang = service.get_language_def("English")["language"]

    assert lang.name == "English"
    assert lang.show_romanization is False, "uses default"
    assert lang.right_to_left is False, "uses default"

    expected = [
        "terms; embeddedhtml; https://en.thefreedictionary.com/###; True; 1",
        "terms; popuphtml; https://www.collinsdictionary.com/dictionary/english/###; True; 2",
        "sentences; popuphtml; https://www.deepl.com/translator#en/en/###; True; 3",
    ]
    actual = [
        f"{ld.usefor}; {ld.dicttype}; {ld.dicturi}; {ld.is_active}; {ld.sort_order}"
        for ld in lang.dictionaries
    ]
    assert actual == expected, "dictionaries"


def test_load_def_loads_lang_and_stories(empty_db):
    "Can load a language."
    story_sql = "select bktitle from books order by BkTitle"
    lang_sql = "select LgName from languages"
    assert_sql_result(lang_sql, [], "no langs")
    assert_sql_result(story_sql, [], "nothing loaded")

    lang_id = service.load_language_def("English")
    assert lang_id > 0, "ID returned, used for filtering"
    assert_sql_result(lang_sql, ["English"], "eng loaded")
    assert_sql_result(story_sql, ["Tutorial", "Tutorial follow-up"], "stories loaded")


def test_load_all_defs_loads_lang_and_stories(empty_db):
    "Smoke test, load everything."
    story_sql = "select bktitle from books"
    lang_sql = "select LgName from languages"
    assert_sql_result(lang_sql, [], "no langs")
    assert_sql_result(story_sql, [], "nothing loaded")

    defs = service.get_defs()
    langnames = [d["language"].name for d in defs]
    for n in langnames:
        lang_id = service.load_language_def(n)
        assert lang_id > 0, "Loaded"



"""
Step defs for term_import.feature.
"""
# pylint: disable=missing-function-docstring

import os
import tempfile
import pytest

from pytest_bdd import given, when, then, scenarios, parsers

from lute.models.language import Language
from lute.models.term import Term

from lute.termimport.service import import_file, BadImportFileError

from tests.dbasserts import assert_sql_result


# The content of the file for the current test.
content = None

# The results of the import
stats = None

scenarios("term_import.feature")
scenarios("term_import_status_0.feature")


@given("demo data")
def given_demo_data(app_context):
    "Calling app_context loads the demo data."


@given(parsers.parse("import file:\n{newcontent}"))
def given_file(newcontent):
    global content  # pylint: disable=global-statement
    content = newcontent


@given("empty import file")
def given_empty_file():
    global content  # pylint: disable=global-statement
    content = ""


@when(parsers.parse("import with create {create}, update {update}"))
def import_with_settings(create, update):
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)

    global stats  # pylint: disable=global-statement
    stats = import_file(path, create.lower() == "true", update.lower() == "true")
    os.remove(path)


@when(
    parsers.parse(
        "import with create {create}, update {update}, new as unknown {newunknowns}"
    )
)
def import_with_settings_and_newunks(create, update, newunknowns):
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)

    global stats  # pylint: disable=global-statement
    stats = import_file(
        path,
        create.lower() == "true",
        update.lower() == "true",
        newunknowns.lower() == "true",
    )
    os.remove(path)


@then(
    parsers.parse(
        "import should succeed with {created} created, {updated} updated, {skipped} skipped"
    )
)
def succeed_with_status(created, updated, skipped):
    assert stats["created"] == int(created), "created"
    assert stats["updated"] == int(updated), "updated"
    assert stats["skipped"] == int(skipped), "skipped"


@then(parsers.parse("import should fail with message:\n{message}"))
def fail_with_message(message):
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)
    with pytest.raises(BadImportFileError, match=message):
        import_file(path)
    os.remove(path)


@then(parsers.parse("words table should contain:\n{text_lc_content}"))
def then_words_table_contains_WoTextLC(text_lc_content):
    expected = []
    if text_lc_content != "-":
        expected = text_lc_content.split("\n")
    sql = "select WoTextLC from words order by WoTextLC"
    assert_sql_result(sql, expected)


@then(parsers.parse('{language} term "{term}" should be:\n{expected}'))
def then_term_tags(language, term, expected):
    lang = Language.find_by_name(language)
    spec = Term(lang, term)
    t = Term.find_by_spec(spec)
    pstring = ", ".join([p.text for p in t.parents])
    if pstring == "":
        pstring = "-"
    tstring = ", ".join([p.text for p in t.term_tags])
    if tstring == "":
        tstring = "-"
    actual = [
        f"translation: {t.translation or '-'}",
        f"pronunciation: {t.romanization or '-'}",
        f"status: {t.status}",
        f"parents: {pstring}",
        f"tags: {tstring}",
    ]
    assert "\n".join(actual) == expected, "term"


@then(parsers.parse('sql "{sql}" should return:\n{expected}'))
def then_sql_returns(sql, expected):
    expected = expected.split("\n")
    assert_sql_result(sql, expected)



"""
Book tests.
"""

import pytest
from lute.term.datatables import get_data_tables_list


@pytest.fixture(name="_dt_params")
def fixture_dt_params():
    "Sample query params."
    columns = [
        {"data": "0", "name": "WoID", "searchable": False, "orderable": False},
        {"data": "1", "name": "WoText", "searchable": True, "orderable": True},
    ]
    params = {
        "draw": "1",
        "columns": columns,
        "order": [{"column": "1", "dir": "asc"}],
        "start": "1",
        "length": "10",
        "search": {"value": "", "regex": False},
        # Filters - set "manually" in the route.
        # Cheating here ... had to look at the request payload
        # in devtools to see what was being sent.
        "filtLanguage": "null",  # Ha!
        "filtParentsOnly": "false",
        "filtAgeMin": "",
        "filtAgeMax": "",
        "filtStatusMin": "0",
        "filtStatusMax": "0",
        "filtIncludeIgnored": "false",
    }
    return params


def test_smoke_term_datatables_query_runs(app_context, _dt_params):
    """
    Smoke test only, ensure query runs.
    """
    get_data_tables_list(_dt_params)
    # print(d['data'])
    a = 1
    assert a == 1, "dummy check"


def test_smoke_query_with_filter_params_runs(app_context, _dt_params):
    "Smoke test with filters set."
    _dt_params["filtLanguage"] = "44"
    _dt_params["filtParentsOnly"] = "true"
    _dt_params["filtAgeMin"] = "1"
    _dt_params["filtAgeMax"] = "10"
    _dt_params["filtStatusMin"] = "2"
    _dt_params["filtStatusMax"] = "4"
    _dt_params["filtIncludeIgnored"] = "true"
    get_data_tables_list(_dt_params)



"Language helper methods."

import os
import re
from glob import glob
import yaml
from lute.models.language import Language
from lute.book.model import Book, Repository
from lute.db import db


def get_defs():
    "Return language definitions."
    ret = []
    def_glob = os.path.join(_language_defs_path(), "**", "definition.yaml")
    for f in glob(def_glob):
        entry = {}
        d = {}
        with open(f, "r", encoding="utf-8") as df:
            d = yaml.safe_load(df)
        lang = Language.from_dict(d)
        entry["language"] = lang
        entry["books"] = _get_books(f, lang.name)
        ret.append(entry)
    ret.sort(key=lambda x: x["language"].name)
    return ret


def predefined_languages():
    "Languages defined in yaml files."
    return [d["language"] for d in get_defs()]


def _get_books(lang_definition_filename, lang_name):
    "Get the stories in the same directory as the definition.yaml."
    books = []
    d, f = os.path.split(lang_definition_filename)
    story_glob = os.path.join(d, "*.txt")
    for filename in glob(story_glob):
        with open(filename, "r", encoding="utf-8") as f:
            content = f.read()
        title_match = re.search(r"title:\s*(.*)\n", content)
        title = title_match.group(1).strip()
        content = re.sub(r"#.*\n", "", content)
        b = Book()
        b.language_name = lang_name
        b.title = title
        b.text = content
        books.append(b)
    return books


def get_language_def(lang_name):
    "Get a lang def and its stories."
    defs = get_defs()
    ret = [d for d in defs if d["language"].name == lang_name]
    if len(ret) == 0:
        raise RuntimeError(f"Missing language def name {lang_name}")
    return ret[0]


def load_language_def(lang_name):
    "Load a language def and its stories, save to database."
    load_def = get_language_def(lang_name)
    lang = load_def["language"]
    db.session.add(lang)
    db.session.commit()

    r = Repository(db)
    for b in load_def["books"]:
        r.add(b)
    r.commit()

    return lang.id


def _language_defs_path():
    "Path to the definitions and stories."
    thisdir = os.path.dirname(__file__)
    d = os.path.join(thisdir, "..", "db", "language_defs")
    return os.path.abspath(d)



"""
Settings routes.
"""

import os
from flask import (
    Blueprint,
    current_app,
    request,
    render_template,
    redirect,
    flash,
    jsonify,
)
from flask_wtf import FlaskForm
from wtforms import BooleanField, StringField, IntegerField, TextAreaField, SelectField
from wtforms.validators import InputRequired, NumberRange
from wtforms import ValidationError
from lute.models.language import Language
from lute.models.setting import UserSetting
from lute.themes.service import list_themes
from lute.db import db
from lute.parse.mecab_parser import JapaneseParser


class UserSettingsForm(FlaskForm):
    """
    Settings.

    Note the field names here must match the keys in the settings table.
    """

    backup_enabled = BooleanField("Backup Enabled")
    backup_dir = StringField("Backup directory")
    backup_auto = BooleanField("Run backups automatically (daily)")
    backup_warn = BooleanField("Warn if backup hasn't run in a week")
    backup_count = IntegerField(
        "Retain backup count",
        validators=[InputRequired(), NumberRange(min=1)],
        render_kw={
            "title": "Count of zipfiles to retain, oldest files are deleted first"
        },
    )

    current_theme = SelectField("Theme", choices=list_themes())
    custom_styles = TextAreaField("Custom styles")
    show_highlights = BooleanField("Highlight terms by status")

    open_popup_in_new_tab = BooleanField("Open popup in new tab")
    stop_audio_on_term_form_open = BooleanField("Stop audio on term form open")

    mecab_path = StringField("MECAB_PATH environment variable")
    reading_choices = [
        ("katakana", "Katakana"),
        ("hiragana", "Hiragana"),
        ("alphabet", "Romaji"),
    ]
    japanese_reading = SelectField("Pronunciation characters", choices=reading_choices)

    def validate_backup_dir(self, field):
        "Field must be set if enabled."
        if self.backup_enabled.data is False:
            return
        v = field.data
        if (v or "").strip() == "":
            raise ValidationError("Backup directory required")

        abspath = os.path.abspath(v)
        if v != abspath:
            msg = f'Backup dir must be absolute path.  Did you mean "{abspath}"?'
            raise ValidationError(msg)
        if not os.path.exists(v):
            raise ValidationError(f'Directory "{v}" does not exist.')
        if not os.path.isdir(v):
            raise ValidationError(f'"{v}" is not a directory.')


bp = Blueprint("settings", __name__, url_prefix="/settings")


@bp.route("/index", methods=["GET", "POST"])
def edit_settings():
    "Edit settings."
    form = UserSettingsForm()

    ac = current_app.env_config
    if ac.is_docker:
        # User shouldn't change some things with docker.
        kw = {"readonly": True, "style": "background-color: LightGray"}
        # Backup dir gets mounted from host.
        form.backup_dir.render_kw = kw

    if form.validate_on_submit():
        # Update the settings in the database
        for field in form:
            if field.id not in ("csrf_token", "submit"):
                UserSetting.set_value(field.id, field.data)
        db.session.commit()

        flash("Settings updated", "success")
        return redirect("/")

    # Load current settings from the database
    for field in form:
        if field.id != "csrf_token":
            field.data = UserSetting.get_value(field.id)
        if isinstance(field, BooleanField):
            # Hack: set boolean settings to ints, otherwise they're always checked.
            field.data = int(field.data or 0)

    return render_template("settings/form.html", form=form)


@bp.route("/test_mecab", methods=["GET"])
def test_parse():
    """
    Do a test parse for the JapaneseParser using the
    given path string.

    Returns { 'success': tokens }, or { 'error' msg }

    """
    mecab_path = request.args.get("mecab_path", None)
    old_setting = UserSetting.get_value("mecab_path")
    result = {"failure": "tbd"}
    try:
        UserSetting.set_value("mecab_path", mecab_path)
        # Parsing requires a language, even if it's a dummy.
        lang = Language()
        p = JapaneseParser()
        src = "私は元気です"
        toks = p.get_parsed_tokens(src, lang)
        toks = [tok.token for tok in toks if tok.token != "¶"]
        message = f"{src} parsed to [{ ', '.join(toks) }]"
        result = {"result": "success", "message": message}
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"{type(e).__name__}: { str(e) }"
        result = {"result": "failure", "message": message}
    finally:
        UserSetting.set_value("mecab_path", old_setting)

    return jsonify(result)


@bp.route("/set/<key>/<value>", methods=["POST"])
def set_key_value(key, value):
    "Set a UserSetting key to value."
    old_value = UserSetting.get_value(key)
    try:
        UserSetting.set_value(key, value)
        result = {"result": "success", "message": "OK"}
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"{type(e).__name__}: { str(e) }"
        UserSetting.set_value(key, old_value)
        result = {"result": "failure", "message": message}
    db.session.commit()
    return jsonify(result)



"""
book helper routines.
"""

import os
from io import StringIO
from datetime import datetime

# pylint: disable=unused-import
from tempfile import TemporaryFile, SpooledTemporaryFile
import requests
from bs4 import BeautifulSoup
from flask import current_app, flash
from openepub import Epub, EpubError
from pypdf import PdfReader
from subtitle_parser import SrtParser, WebVttParser
from werkzeug.utils import secure_filename
from lute.book.model import Book


class BookImportException(Exception):
    """
    Exception to throw on book import error.
    """

    def __init__(self, message="A custom error occurred", cause=None):
        self.cause = cause
        self.message = message
        super().__init__(message)


def _secure_unique_fname(filename):
    """
    Return secure name pre-pended with datetime string.
    """
    current_datetime = datetime.now()
    formatted_datetime = current_datetime.strftime("%Y%m%d_%H%M%S")
    f = "_".join([formatted_datetime, secure_filename(filename)])
    return f


def save_audio_file(audio_file_field_data):
    """
    Save the file to disk, return its filename.
    """
    filename = _secure_unique_fname(audio_file_field_data.filename)
    fp = os.path.join(current_app.env_config.useraudiopath, filename)
    audio_file_field_data.save(fp)
    return filename


def get_file_content(filefielddata):
    """
    Get the content of the file.
    """
    content = None
    _, ext = os.path.splitext(filefielddata.filename)
    ext = (ext or "").lower()
    if ext == ".txt":
        content = get_textfile_content(filefielddata)
    if ext == ".epub":
        content = get_epub_content(filefielddata)
    if ext == ".pdf":
        msg = """
        Note: pdf imports can be inaccurate, due to how PDFs are encoded.
        Please be aware of this while reading.
        """
        flash(msg, "notice")
        content = get_pdf_content_from_form(filefielddata)
    if ext == ".srt":
        content = get_srt_content(filefielddata)
    if ext == ".vtt":
        content = get_vtt_content(filefielddata)

    if content is None:
        raise ValueError(f'Unknown file extension "{ext}"')
    if content.strip() == "":
        raise BookImportException(f"{filefielddata.filename} is empty.")
    return content


def get_textfile_content(filefielddata):
    "Get content as a single string."
    content = ""
    try:
        content = filefielddata.read()
        return str(content, "utf-8")
    except UnicodeDecodeError as e:
        f = filefielddata.filename
        msg = f"{f} is not utf-8 encoding, please convert it to utf-8 first (error: {str(e)})"
        raise BookImportException(message=msg, cause=e) from e


def get_epub_content(epub_file_field_data):
    """
    Get the content of the epub as a single string.
    """
    content = ""
    try:
        if hasattr(epub_file_field_data.stream, "seekable"):
            epub = Epub(stream=epub_file_field_data.stream)
            content = epub.get_text()
        else:
            # We get a SpooledTemporaryFile from the form but this doesn't
            # implement all file-like methods until python 3.11. So we need
            # to rewrite it into a TemporaryFile
            with TemporaryFile() as tf:
                epub_file_field_data.stream.seek(0)
                tf.write(epub_file_field_data.stream.read())
                epub = Epub(stream=tf)
                content = epub.get_text()
    except EpubError as e:
        msg = f"Could not parse {epub_file_field_data.filename} (error: {str(e)})"
        raise BookImportException(message=msg, cause=e) from e
    return content


def get_pdf_content_from_form(pdf_file_field_data):
    "Get content as a single string from a PDF file using PyPDF2."
    content = ""
    try:
        pdf_reader = PdfReader(pdf_file_field_data)

        for page in pdf_reader.pages:
            content += page.extract_text()

        return content
    except Exception as e:
        msg = f"Could not parse {pdf_file_field_data.filename} (error: {str(e)})"
        raise BookImportException(message=msg, cause=e) from e


def get_srt_content(srt_file_field_data):
    """
    Get the content of the srt as a single string.
    """
    content = ""
    try:
        srt_content = srt_file_field_data.read().decode("utf-8-sig")

        parser = SrtParser(StringIO(srt_content))
        parser.parse()

        content = "\n".join(subtitle.text for subtitle in parser.subtitles)

        return content
    except Exception as e:
        msg = f"Could not parse {srt_file_field_data.filename} (error: {str(e)})"
        raise BookImportException(message=msg, cause=e) from e


def get_vtt_content(vtt_file_field_data):
    """
    Get the content of the vtt as a single string.
    """
    content = ""
    try:
        vtt_content = vtt_file_field_data.read().decode("utf-8-sig")

        # Check if it is from YouTube
        lines = vtt_content.split("\n")
        if lines[1].startswith("Kind:") and lines[2].startswith("Language:"):
            vtt_content = "\n".join(lines[:1] + lines[3:])

        parser = WebVttParser(StringIO(vtt_content))
        parser.parse()

        content = "\n".join(subtitle.text for subtitle in parser.subtitles)

        return content
    except Exception as e:
        msg = f"Could not parse {vtt_file_field_data.filename} (error: {str(e)})"
        raise BookImportException(message=msg, cause=e) from e


def book_from_url(url):
    "Parse the url and load a new Book."
    s = None
    try:
        timeout = 20  # seconds
        response = requests.get(url, timeout=timeout)
        response.raise_for_status()
        s = response.text
    except requests.exceptions.RequestException as e:
        msg = f"Could not parse {url} (error: {str(e)})"
        raise BookImportException(message=msg, cause=e) from e

    soup = BeautifulSoup(s, "html.parser")
    extracted_text = []

    # Add elements in order found.
    for element in soup.descendants:
        if element.name in ("h1", "h2", "h3", "h4", "p"):
            extracted_text.append(element.text)

    title_node = soup.find("title")
    orig_title = title_node.string if title_node else url

    short_title = orig_title[:150]
    if len(orig_title) > 150:
        short_title += " ..."

    b = Book()
    b.title = short_title
    b.source_uri = url
    b.text = "\n\n".join(extracted_text)
    return b



"""
Term mapping tests.
"""

import pytest
from lute.models.term import Term, TermTag, TermTextChangedException
from lute.db import db
from tests.dbasserts import assert_sql_result


def test_save_and_remove(empty_db, english):
    "Smoke test of mappings and db records."
    sql = "select WoText, WoTextLC, WoTokenCount from words"
    assert_sql_result(sql, [], "empty table")

    term = Term(english, "ABC")

    db.session.add(term)
    db.session.commit()
    assert_sql_result(sql, ["ABC; abc; 1"], "have term")

    db.session.delete(term)
    db.session.commit()
    assert_sql_result(sql, [], "no terms")


def test_term_with_parent_and_tags(empty_db, spanish):
    "Misc data check - parent and tags are saved."
    term = Term(spanish, "HOLA")
    parent = Term(spanish, "PARENT")
    term.add_parent(parent)

    tg = TermTag("tag")
    term.add_term_tag(tg)

    db.session.add(term)
    db.session.commit()

    sql = "select WoID, WoText, WoTextLC from words"
    expected = ["1; HOLA; hola", "2; PARENT; parent"]
    assert_sql_result(sql, expected, "sanity check on save")

    assert_sql_result("select WpWoID, WpParentWoID from wordparents", ["1; 2"], "wp")

    # Hacky sql check.
    sql = """select w.WoText, p.WoText as ptext, tags.TgText
        FROM words w
        INNER JOIN wordparents on WpWoID = w.WoID
        INNER JOIN words p on p.WoID = wordparents.WpParentWoID
        INNER JOIN wordtags on WtWoID = w.WoID
        INNER JOIN tags on TgID = WtTgID"""
    exp = ["HOLA; PARENT; tag"]
    assert_sql_result(sql, exp, "parents, tags")


def test_term_parent_with_two_children(spanish):
    "Both children get saved in the db."
    term = Term(spanish, "HOLA")
    gato = Term(spanish, "gato")
    parent = Term(spanish, "PARENT")
    gato.add_parent(parent)
    term.add_parent(parent)

    db.session.add(term)
    db.session.add(gato)
    db.session.commit()

    parent_get = Term.find(parent.id)
    assert parent_get.text == "PARENT"
    assert len(parent_get.children) == 2

    expected = [f"{term.id}; {parent.id}", f"{gato.id}; {parent.id}"]
    expected.sort()
    sql = """select WpWoID, WpParentWoID
    from wordparents order by WpWoID"""
    assert_sql_result(sql, expected, "wp")


def test_add_and_remove_parents(spanish):
    "Parent mappings."
    term = Term(spanish, "HOLA")
    p1 = Term(spanish, "parent")
    p2 = Term(spanish, "otro")
    term.add_parent(p1)
    term.add_parent(p2)

    db.session.add(term)
    db.session.commit()

    sql = """select WpWoID, WpParentWoID
    from wordparents order by WpWoID, WpParentWoID"""

    expected = [
        f"{term.id}; {p1.id}",
        f"{term.id}; {p2.id}",
    ]
    expected.sort()
    assert_sql_result(sql, expected, "after change")

    term.remove_all_parents()
    db.session.add(term)
    db.session.commit()
    assert_sql_result(sql, [], "all removed")


def test_remove_parent_leaves_children_in_db(spanish):
    "Child deletion doesn't delete the parent."
    term = Term(spanish, "HOLA")
    parent = Term(spanish, "PARENT")
    term.add_parent(parent)

    db.session.add(term)
    db.session.commit()
    sql_list = "select WoText from words order by WoText"
    sql = f"""select w.WoText, p.WoText as ptext from words w
    left join wordparents on WpWoID = w.WoID
    left join words p on p.WoID = wordparents.WpParentWoID
    where w.WoID = {term.id}"""
    assert_sql_result(sql_list, ["HOLA", "PARENT"], "both exist")
    assert_sql_result(sql, ["HOLA; PARENT"], "parent set")

    db.session.delete(parent)
    db.session.commit()
    assert_sql_result(sql_list, ["HOLA"], "parent removed")
    assert_sql_result(sql, ["HOLA; None"], "parent not set")


def test_removing_term_with_parent_and_tag(empty_db, spanish):
    """
    Both the parent and the tag are left.
    """
    term = Term(spanish, "HOLA")
    parent = Term(spanish, "PARENT")
    term.add_parent(parent)
    tg = TermTag("tag")
    term.add_term_tag(tg)

    db.session.add(term)
    db.session.commit()
    sqllist = "select WoText from words order by WoText"
    sqltags = "select TgText from tags"
    assert_sql_result(sqllist, ["HOLA", "PARENT"], "both exist")
    assert_sql_result(sqltags, ["tag"], "tag exists")

    db.session.delete(term)
    db.session.commit()
    assert_sql_result(sqllist, ["PARENT"], "parent left")
    assert_sql_result(sqltags, ["tag"], "tag left")


def test_save_replace_remove_image(spanish):
    "Save saves the associated image."
    t = Term(spanish, "HOLA")
    t.set_current_image("hello.png")
    assert t.get_current_image() == "hello.png"

    db.session.add(t)
    db.session.commit()
    sql = "select WiWoID, WiSource from wordimages"
    expected = ["1; hello.png"]
    assert_sql_result(sql, expected, "hello")

    t.set_current_image("there.png")
    assert t.get_current_image() == "there.png"

    db.session.add(t)
    db.session.commit()
    expected = ["1; there.png"]
    assert_sql_result(sql, expected, "there")

    t.set_current_image(None)
    assert t.get_current_image() is None

    db.session.add(t)
    db.session.commit()
    expected = []
    assert_sql_result(sql, expected, "removed")


def test_delete_term_deletes_image(spanish):
    "Check cascade delete."
    t = Term(spanish, "HOLA")
    t.set_current_image("hello.png")
    assert t.get_current_image() == "hello.png"

    db.session.add(t)
    db.session.commit()
    sql = "select WiWoID, WiSource from wordimages"
    expected = ["1; hello.png"]
    assert_sql_result(sql, expected, "hello")

    db.session.delete(t)
    db.session.commit()
    assert_sql_result(sql, [], "removed")


def test_save_remove_flash_message(spanish):
    "Flash message is associated with term, can be popped."
    t = Term(spanish, "HOLA")
    t.set_flash_message("hello")
    assert t.get_flash_message() == "hello"

    db.session.add(t)
    db.session.commit()
    sql = "select WfWoID, WfMessage from wordflashmessages"
    expected = ["1; hello"]
    assert_sql_result(sql, expected, "hello")

    assert t.pop_flash_message() == "hello", "popped"
    assert t.get_flash_message() is None, "no message now"

    db.session.add(t)
    db.session.commit()
    assert_sql_result(sql, [], "popped")


def test_delete_term_deletes_flash_message(spanish):
    "Check cascade delete."
    t = Term(spanish, "HOLA")
    t.set_flash_message("hello")

    db.session.add(t)
    db.session.commit()
    sql = "select WfWoID, WfMessage from wordflashmessages"
    expected = ["1; hello"]
    assert_sql_result(sql, expected, "hello")

    db.session.delete(t)
    db.session.commit()
    assert_sql_result(sql, [], "popped")


## Changing term text isn't allowed -- changing case is ok.


@pytest.mark.term_case
def test_changing_text_of_saved_Term_throws(english):
    "Changing text should throw."
    term = Term(english, "ABC")
    db.session.add(term)
    db.session.commit()

    with pytest.raises(TermTextChangedException):
        term.text = "DEF"


@pytest.mark.term_case
def test_changing_text_to_same_thing_does_not_throw(japanese):
    """
    New terms get created on page open, and when parsed
    in full context of the text, the term may be parsed differently
    than when the form is opened.

    e.g. for Japanese, the term "集めれ" may get created on
    page load due to the term getting parsed in the full page context,
    but when parsed individually it's parsed as "集め/れ".
    Setting a new term with original text "集めれ" to that _same_ text
    means that the text is really _unchanged_.
    """
    term = Term.create_term_no_parsing(japanese, "集めれ")
    db.session.add(term)
    db.session.commit()

    sql = f"select wotext, wotextlc from words where woid = {term.id}"
    assert_sql_result(sql, ["集めれ; 集めれ"], "have term")

    t = Term.find(term.id)
    t.text = "集めれ"
    db.session.add(t)
    db.session.commit()
    assert_sql_result(sql, ["集めれ; 集めれ"], "have term")


@pytest.mark.term_case
def test_changing_multiword_text_case_does_not_throw(english):
    """
    Sanity check.
    """
    term = Term(english, "a CAT")
    db.session.add(term)
    db.session.commit()

    sql = f"select wotext, wotextlc from words where woid = {term.id}"
    assert_sql_result(sql, ["a/ /CAT; a/ /cat"], "have term")

    t = Term.find(term.id)
    t.text = term.text.lower()
    db.session.add(t)
    db.session.commit()
    assert_sql_result(sql, ["a/ /cat; a/ /cat"], "have term")


@pytest.mark.term_case
def test_changing_case_only_of_text_of_saved_Term_is_ok(english):
    "Changing text should throw."
    term = Term(english, "ABC")
    db.session.add(term)
    db.session.commit()

    term.text = "abc"


@pytest.mark.term_case
def test_changing_text_of_non_saved_Term_is_ok(english):
    "Changing text should not throw if not saved."
    term = Term(english, "ABC")
    term.text = "DEF"



"""
Text mapping checks.
"""

from datetime import datetime
from lute.models.book import Book, Text
from lute.db import db
from tests.dbasserts import assert_record_count_equals


def test_save_text_sentences_replaced_in_db(empty_db, english):
    """
    Sentences should only be generated when a Text is saved with the ReadDate saved.
    Sentences are only used for reference lookups.
    """
    b = Book("hola", english)
    t = Text(b, "Tienes un perro. Un gato.")

    db.session.add(t)
    db.session.commit()
    assert_record_count_equals("sentences", 0, "no sentences")

    t.read_date = datetime.now()
    db.session.add(t)
    db.session.commit()
    assert_record_count_equals("sentences", 2, "2 sentences")

    t.text = "Tengo un coche."
    db.session.add(t)
    db.session.commit()
    assert_record_count_equals("sentences", 1, "back to 1 sentences")



"""
DataTables sqlite tests.
"""

from lute.utils.formutils import language_choices, valid_current_language_id
from lute.models.setting import UserSetting


def test_language_choices(app_context):
    "Gets all languages."
    choices = language_choices()
    assert choices[0][1] == "-", "- at the top"
    langnames = [c[1] for c in choices]
    assert "Spanish" in langnames, "have Spanish"


def test_valid_current_language_id(app_context):
    "Sanity check only."
    UserSetting.set_value("current_language_id", 9999)
    assert int(UserSetting.get_value("current_language_id")) == 9999, "pre-check"
    curr_lang_id = int(valid_current_language_id())
    assert curr_lang_id == 0, "set back to 0"
    assert int(UserSetting.get_value("current_language_id")) == 0, "re-set to 0"



"""
/language endpoints.
"""

from sqlalchemy import func
from sqlalchemy.exc import IntegrityError
from flask import Blueprint, current_app, render_template, redirect, url_for, flash
from lute.models.language import Language
from lute.models.setting import UserSetting
from lute.models.book import Book
from lute.models.term import Term
import lute.language.service
from lute.language.forms import LanguageForm
from lute.db import db
from lute.parse.registry import supported_parsers

bp = Blueprint("language", __name__, url_prefix="/language")


@bp.route("/index")
def index():
    """
    List all languages.

    This includes the Book and Term count for each Language.  These
    counts are pulled in by subqueries, because Language doesn't have
    "books" and "terms" members ... I was having trouble with session
    management when these were added, and they're only used here, so
    this is good enough for now.
    """

    def create_count_subquery(class_, count_column):
        # Re the pylint disable, ref
        # https://github.com/pylint-dev/pylint/issues/8138 ...
        ret = (
            db.session.query(
                class_.language_id,
                # pylint: disable=not-callable
                func.count(class_.id).label(count_column),
            )
            .group_by(class_.language_id)
            .subquery()
        )
        return ret

    # Create subqueries for counting books and terms
    book_subquery = create_count_subquery(Book, "book_count")
    term_subquery = create_count_subquery(Term, "term_count")

    # Query to join Language with book and term counts
    query = (
        db.session.query(
            Language, book_subquery.c.book_count, term_subquery.c.term_count
        )
        .outerjoin(book_subquery, Language.id == book_subquery.c.language_id)
        .outerjoin(term_subquery, Language.id == term_subquery.c.language_id)
    )

    results = query.all()

    results = [rec for rec in results if rec[0].is_supported is True]

    return render_template("language/index.html", language_data=results)


def _handle_form(language, form) -> bool:
    """
    Handle the language form processing.

    Returns True if validated and saved.
    """
    result = False

    if form.validate_on_submit():
        try:
            form.populate_obj(language)
            current_app.db.session.add(language)
            current_app.db.session.commit()
            flash(f"Language {language.name} updated", "success")
            result = True
        except IntegrityError as e:
            msg = e.orig
            if "languages.LgName" in f"{e.orig}":
                msg = f"{language.name} already exists."
            flash(msg, "error")

    return result


def _add_hidden_dictionary_template_entry(form):
    "Add a dummy placeholder dictionary to be used as a template."
    # Add a dummy dictionary entry with dicturi __TEMPLATE__.
    #
    # This entry is used as a "template" when adding a new dictionary
    # to the list of dictionaries (see templates/language/_form.html).
    # This is the easiest way to ensure that new dictionary entries
    # have the correct controls.
    #
    # This dummy entry is not rendered on the form, or submitted
    # when the form is submitted.  Search for __TEMPLATE__ in
    # templates/language/_form.html to see where it is handled.
    form.dictionaries.append_entry({"dicturi": "__TEMPLATE__"})


@bp.route("/edit/<int:langid>", methods=["GET", "POST"])
def edit(langid):
    """
    Edit a language.
    """
    language = db.session.get(Language, langid)

    if not language:
        flash(f"Language {langid} not found", "danger")
        return redirect(url_for("language.index"))

    form = LanguageForm(obj=language)
    form.parser_type.choices = supported_parsers()

    if _handle_form(language, form):
        return redirect("/")

    _add_hidden_dictionary_template_entry(form)

    return render_template("language/edit.html", form=form, language=language)


@bp.route("/new", defaults={"langname": None}, methods=["GET", "POST"])
@bp.route("/new/<string:langname>", methods=["GET", "POST"])
def new(langname):
    """
    Create a new language.
    """
    predefined = lute.language.service.predefined_languages()
    language = Language()
    if langname is not None:
        candidates = [lang for lang in predefined if lang.name == langname]
        if len(candidates) == 1:
            language = candidates[0]

    form = LanguageForm(obj=language)
    form.parser_type.choices = supported_parsers()

    if _handle_form(language, form):
        # New language, so show everything b/c user should re-choose
        # the default.
        #
        # Reason for this: a user may start off with just language X,
        # so the current_language_id is set to X.id.  If the user then
        # adds language Y, the filter stays on X, which may be
        # disconcerting/confusing.  Forcing a reselect is painless and
        # unambiguous.
        UserSetting.set_value("current_language_id", 0)
        db.session.commit()
        return redirect("/")

    _add_hidden_dictionary_template_entry(form)

    return render_template(
        "language/new.html", form=form, language=language, predefined=predefined
    )


@bp.route("/delete/<int:langid>", methods=["POST"])
def delete(langid):
    """
    Delete a language.
    """
    language = db.session.get(Language, langid)
    if not language:
        flash(f"Language {langid} not found")
    Language.delete(language)
    return redirect(url_for("language.index"))


@bp.route("/list_predefined", methods=["GET"])
def list_predefined():
    "Show predefined languages that are not already in the db."
    predefined = lute.language.service.predefined_languages()
    existing_langs = db.session.query(Language).all()
    existing_names = [l.name for l in existing_langs]
    new_langs = [p for p in predefined if p.name not in existing_names]
    return render_template("language/list_predefined.html", predefined=new_langs)


@bp.route("/load_predefined/<langname>", methods=["GET"])
def load_predefined(langname):
    "Load a predefined language and its stories."
    lang_id = lute.language.service.load_language_def(langname)
    UserSetting.set_value("current_language_id", lang_id)
    db.session.commit()
    flash(f"Loaded {langname} and sample book(s)")
    return redirect("/")
